{-# OPTIONS_GHC -Wall -fno-warn-orphans #-}
{-# LANGUAGE ScopedTypeVariables #-}

-- Testing
module HarmTrace.IO.Main where

-- Parser stuff
import Text.ParserCombinators.UU

-- Music stuff
import HarmTrace.HarmTrace
import HarmTrace.Base.MusicRep
import HarmTrace.Model.Instances ()
import HarmTrace.HAnTree.Tree (Tree)
import HarmTrace.HAnTree.ToHAnTree
import HarmTrace.IO.Errors

-- Library modules
import System.Console.ParseArgs hiding (args) -- cabal install parseargs
import Data.List (sort, groupBy, intersperse)
import Data.Binary (encodeFile)
import Control.Arrow ((***))
import System.FilePath
import System.Directory
import System.IO
import Text.Regex.TDFA
import Text.Printf (printf)
import System.CPUTime


--------------------------------------------------------------------------------
-- Data set Info
--------------------------------------------------------------------------------
biabPat :: String
biabPat = "^(.*)_id_([0-9]{5})_(allanah|wdick|community|midicons|realbook).(M|S|m|s)(G|g)[0-9A-Za-z]{1}.txt$"     
    
getInfo :: String -> Maybe [String]     
getInfo fileName = 
  do let 
     (_,_,_,groups) <- fileName =~~ biabPat :: Maybe (String,String,String,[String])
     return groups

getTitle, getId, getDb :: String -> String
getTitle fn = getInfo' 0 fn     
getId fn    = getInfo' 1 fn
getDb fn    = getInfo' 2 fn
    
getInfo' :: Int -> String -> String    
getInfo' i fn = maybe "no_info" (!!i) (getInfo fn)
                    
createGroundTruth :: [String] -> [(String, String)]
createGroundTruth files = [ (getTitle x, getId x) | x <- files ]

getClassSizes :: [String] -> [(String,[String])]
getClassSizes = map ((head *** id) . unzip) . groupBy gf . createGroundTruth
  where gf (name1, _key1) (name2, _key2) = name1 == name2

writeGroundTruth :: [FilePath] -> FilePath -> IO ()
writeGroundTruth files outfp =
  writeFile outfp . Prelude.tail $ concatMap merge (createGroundTruth files) 
     where merge :: (String, String) -> String
           merge (x,y) = '\n' : y ++ "\t" ++ x

--------------------------------------------------------------------------------
-- Symbolic IO
--------------------------------------------------------------------------------

-- parses a string of chords s and returns a parse tree with the harmony structure
parseTree, parseTreeVerb :: String -> IO [Tree HAn]
parseTree s = do let (toks, _, ts, te, pe) = string2PieceCPostProc s
                 putStrLn ("parsed " ++ show (length toks) ++ " chords")
                 if not $ null te then showErrors "tokenizer: " te 
                                  else putStr ""
                 if not $ null pe then showErrors "parser: " pe 
                                  else putStr ""
                 return ts

parseTreeVerb s = 
  do let (toks, _, ts, te, pe) = string2PieceCPostProc s
     putStrLn ("parsed " ++ show (length toks) ++ " chords")
     if not (null te) then mapM_ print te
                      else putStr ""
     if not (null pe) then mapM_ print pe
                      else putStr ""
     return ts                 
      
-- Batch analyzing a directory with chord sequence files with reduced output.
parseDir :: FilePath -> Maybe FilePath -> IO ()
parseDir filepath bOut =     getDirectoryContents filepath 
                         >>= parseDir' bOut filepath . sort

parseDir' :: Maybe FilePath -> String -> [String] -> IO ()
parseDir' bOut fp fs = 
  do putStr "Filename\tNumber of trees\t"
     putStr "Insertions\tDeletions\tReplacements\tDeletions at the end\t"
     putStr "Tot_Correction\tNr_of_chords\t"
     putStrLn "Error ratio\tTime taken"
     let process :: FilePath -> FilePath -> IO ([ChordLabel],Tree HAn)
         process path x =
          do content <- readFile (path </> x)
             let (tks, ps, ts, e1, e2) = string2PieceCPostProc content
                 t                = seq (length ts) (return ())
                 ErrorNrs i d e r = countErrors e2
                 errRat           = errorRatio e2 tks
                 nrOfChords       = length tks -- (mergeDups toks)
             t1 <- getCPUTime
             t
             t2 <- getCPUTime
             let diff = fromIntegral (t2 - t1) / (1000000000 :: Float)
             when (not (null e1)) $
               putErrLn (show x ++ ": " ++ show e1)
             printLn . concat $ intersperse "\t" [ x, show (length ps)
                                       , show i, show d, show r, show e
                                       , show (i+d+e+r)
                                       , show nrOfChords, showFloat errRat
                                       , showFloat diff]   
             return (tks, head ts)
     res <- mapM (process fp) (filter ((== ".txt") . takeExtension) fs)
     case bOut of
       Nothing -> return ()
       Just bf -> encodeFile bf (unzip res :: ([[ChordLabel]],[Tree HAn]))
      
--------------------------------------------------------------------------------
-- Utils
--------------------------------------------------------------------------------
putErrLn :: String -> IO()
putErrLn = hPutStrLn stderr

printLn :: String ->IO ()
printLn s = putStrLn s >> hFlush stdout


regexMatchGroups :: String -> String -> Maybe [String]
regexMatchGroups pat str = do
  (_,_,_,groups) <- str =~~ pat :: Maybe (String,String,String,[String])
  return groups

-- Stricter readFile
hGetContents' :: Handle -> IO [Char]
hGetContents' hdl = do e <- hIsEOF hdl
                       if e then return []
                         else do c <- hGetChar hdl
                                 cs <- hGetContents' hdl
                                 return (c:cs)

readFile' :: FilePath -> IO [Char]
readFile' fn = do hdl <- openFile fn ReadMode
                  xs <- hGetContents' hdl
                  hClose hdl
                  return xs
                  
readDataDir :: FilePath -> IO [FilePath]
readDataDir fp = 
  do fs <- getDirectoryContents fp
     return . sort $ filter (\str -> str =~ biabPat) fs

-- | Shows a Float with 5 decimal places
showFloat :: Float -> String
showFloat = printf "%.6f"
--showFloat =   show . (/ (1000 :: Float)) . fromIntegral 
  --          . (round :: Float -> Int) . (* 1000)
      
 
