{-# OPTIONS_GHC -Wall #-}
module HarmTrace.Matching.GuptaNishimuraEditMatch ( getSimLCES, getLCES
                                            , getWeightLCES
                                            )where


--------------------------------------------------------------------------------
-- Finding the Largest Common Embedable Subtrees (LCES)
-- Based on: Gupta, A. and Nishimura, N. (1998) Finding largest subtrees and 
--           smallest supertrees, Algorithmica, 21(2), p. 183--210
-- author: Bas de Haas
-------------------------------------------------------------------------------- 

import qualified Data.Vector as V

import Data.Ord
import Data.Maybe
import Data.Array
import Data.List

import HarmTrace.HAnTree.Tree
import HarmTrace.Matching.Sim

-------------------------------------------------------------------------------- 
-- Top Level LCES function
-------------------------------------------------------------------------------- 

getLCES :: (Eq t, Sim t) => Tree t -> Tree t -> [Tree t] 
getLCES ta tb = fst (getWeightLCES ta  tb)

getSimLCES :: (Sim t, Eq t) => Tree t -> Tree t -> Float
getSimLCES ta tb = (weight * weight) / (selfSim ta * selfSim tb) where
     (_lces, weight) = getWeightLCES ta tb

  
-- Top level function that returns the largest common embedable subtree
-- of two trees
getWeightLCES :: (Eq t, Sim t) => Tree t -> Tree t -> ([Tree t],Float)
getWeightLCES ta tb = (matchToTree ta (map fst (reverse m)),s) where 
  n     = lces  ta tb
  (m,s) = n!(snd $ bounds n)
  
 
-- calculates the largest labeled common embeddable subtree
lces :: (Eq t, Sim t) => Tree t -> Tree t 
     -> Array (Int, Int) ([(Int,Int)], Float)
lces ta tb = n where
  la = size ta-1
  lb = size tb-1
  a = V.fromList (pot ta)
  b = V.fromList (pot tb)
  maxi :: Int -> [Int] -> ([(Int,Int)],Float)
  maxi _ [] = ([],0)  
  maxi i cb = {-# SCC "maxi_lces" #-}n!(i,maximumBy (comparing (\j -> getWeight $ n!(i,j))) cb  )
  maxj :: [Int] -> Int -> ([(Int,Int)],Float)
  maxj [] _ = ([],0)  
  maxj ca j = {-# SCC "maxi_lces" #-}n!(  maximumBy (comparing (\i -> getWeight $ n!(i,j))) ca,j)
  recur i j = findBestMatch (sim (getLabel $ a V.! i) (getLabel $ b V.! j))i j mc mi mj where  
    mi  = maxi i  (getChildPns (b V.! j))
    mj  = maxj    (getChildPns (a V.! i)) j
    mc  = wbMatch (getChild (a V.! i)) (getChild $ b V.! j) n
  n = array ((0,0), (la, lb))     
      (((0,0), ([],sim (getLabel $ a V.! 0) (getLabel (b V.! 0)))) :
      [((0,j), recur 0 j) | j <- [1..lb]] ++
      [((i,0), recur i 0) | i <- [1..la]] ++
      [((i,j), recur i j) | i <- [1..la], j <- [1..lb]])
 
-- returns the best matching candidate, given the previous candidates, the
-- bipartite matching. The function depends on wheter the currend nodes 
-- match and wether, in that case, on of the current nodes is not allready 
-- matched
findBestMatch :: Float -> Int -> Int -> ([(Int,Int)], Float) 
  -> ([(Int,Int)], Float) -> ([(Int,Int)], Float) -> ([(Int,Int)], Float) 
findBestMatch simv i j a b c
  | simv == 0 = first
  | otherwise = if isFree first i j       then ((i,j):mf,wf+simv) --add match 
                else if wf /= ws          then first           
                else if isFree second i j then ((i,j):ms,ws+simv)
                else if wf /= wt          then first
                else if isFree second i j then ((i,j):mt,wt+simv)
                else first where
    (first@(mf,wf) : second@(ms,ws) : (mt,wt) : []) = 
       {- SCC sorting -} reverse $ sortBy (comparing getWeight) [a,b,c]   
       
          
--------------------------------------------------------------------------------
-- Weighted Plannar Matching of a Bipartite Graph
--------------------------------------------------------------------------------   
      
-- selects the most lower right cell in the wbMatch' matrix
wbMatch  :: (Eq t) => [Tree t] -> [Tree t] 
         -> Array (Int, Int) ([(Int, Int)], Float) -> ([(Int, Int)], Float)
wbMatch  _ []  _ = ([],0)
wbMatch  [] _  _ = ([],0)
wbMatch  a  b  n = getDownRight $ wbMatch' a b n


-- returns the actual planar weighted bipartite matchings. n should contain 
-- the weights of the edge between a[i] and b[j]  
wbMatch'   :: (Eq t) => [Tree t] -> [Tree t] 
           ->  Array (Int, Int) ([(Int,Int)], Float)
           ->  Array (Int, Int) ([(Int,Int)], Float)
wbMatch' _ []  _ = {-# SCC "listArrayA" #-} listArray ((0,0),(0,0)) []
wbMatch' [] _  _ = {-# SCC "listArrayB" #-} listArray ((0,0),(0,0)) []
wbMatch' a b n = m where
  la = length a-1
  lb = length b-1   
  -- returns a previously matched subtree 
  subTree :: Int -> Int -> ([(Int,Int)], Float) 
  subTree i j = n ! (fromJust . getPn $ a!!i, fromJust . getPn $ b!!j)
  -- this is the actual core recursive definintion of the algorithm
  match :: Int -> Int -> ([(Int,Int)], Float) 
  match i j = maximumBy (comparing getWeight) [maxPrv, minPrv, diagM] where
    s@(_mat,w) = subTree i j
    hasMatch  = w > 0
    maxPrv    = if not hasMatch                then m!(i-1,j)
                else if isFree (m!(i-1,j)) i j then merge s (m!(i-1,j))
                else m!(i-1,j)
    minPrv    = if not hasMatch                then m!(i,j-1)
                else if isFree (m!(i,j-1)) i j then merge s (m!(i,j-1)) 
                else m!(i,j-1)
    diagM     = merge s (m!(i-1,j-1)) 
  m = array ((0,0),(la,lb)) 
    (((0,0), subTree 0 0) :
    [((0,j), if getWeight (subTree 0 j) > getWeight (m!(0,j-1)) 
             then subTree 0 j else m!(0,j-1)) | j <- [1..lb]] ++
    [((i,0), if getWeight (subTree i 0) > getWeight (m!(i-1,0)) 
             then subTree i 0 else m!(i-1,0)) | i <- [1..la]] ++
    [((i,j), match i j) | i <- [1..la], j <- [1..lb]])     

     
--------------------------------------------------------------------------------
-- Some LCES helper functions
-------------------------------------------------------------------------------- 
getDownRight :: (Ix i) => Array i e -> e
getDownRight n = n ! snd (bounds n)

-- returns the weight of a match and is synonymous to snd
getWeight :: (a, b) -> b
getWeight (_,w) = w

-- checks if the previously calculated optimal solution does not 
-- contain the indices i and j in a and b, resepectivly
isFree :: ([(Int,Int)], Float) -> Int -> Int -> Bool
isFree ([],_)               _ _ = True
isFree ((previ, prevj):_,_) i j = ( i > previ && j > prevj)

-- mergest two lists with matches
merge :: ([a], Float) -> ([a], Float) -> ([a], Float)
merge (a, wa) (b, wb) = (a ++ b, wa + wb)
